#### ロジスティック回帰による手書き文字認識
import os
import struct
import matplotlib.pyplot as plt
import numpy as np
import sys
from sklearn.linear_model import LogisticRegression

# MNISTデータの読み込み関数
def load_mnist(path, kind='train'):

    labels_path = os.path.join(path,'%s-labels-idx1-ubyte'% kind)
    images_path = os.path.join(path,'%s-images-idx3-ubyte'% kind)

    with open(labels_path, 'rb') as lbpath:
        magic, n = struct.unpack('>II',lbpath.read(8))
        labels = np.fromfile(lbpath,dtype=np.uint8)

    with open(images_path, 'rb') as imgpath:
        magic, num, rows, cols = struct.unpack(">IIII",imgpath.read(16))
        images = np.fromfile(imgpath,dtype=np.uint8).reshape(len(labels), 784)

    return images, labels

# MNISTデータの読み込み
current_path = os.path.dirname(os.path.realpath("__file__"))
X_train, y_train = load_mnist(current_path, kind='train')
X_test, y_test = load_mnist(current_path, kind='t10k')

# 学習用に最初の1000点，テスト用に最初の300点のデータを使用
X_train = X_train[:1000][:]
y_train = y_train[:1000][:]
X_test = X_test[:300][:]
y_test = y_test[:300][:]
print('#data: %d, #feature: %d (training data)' % (X_train.shape[0], X_train.shape[1]))
print('#data: %d, #feature: %d (test data)' % (X_test.shape[0], X_test.shape[1]))

# ロジスティック回帰のインスタンスの生成と学習
lr = LogisticRegression(penalty='l1', C=1000.0, random_state=0)
lr.fit(X_train, y_train)

# 学習データおよびテストデータに対するaccuracyの算出
y_train_pred = lr.predict(X_train)
acc = np.sum(y_train == y_train_pred, axis=0)*100 / X_train.shape[0]
print('accuracy for training data: %.2f%%' % acc)

y_test_pred = lr.predict(X_test)
acc = np.sum(y_test == y_test_pred, axis=0)*100 / X_test.shape[0]
print('accuracy for test data: %.2f%%' % acc)

# 最初の25サンプルの識別結果をプロット．t: 正解クラス，p: 識別器による推測クラス
orign_img = X_test[:25][:25]
true_lab = y_test[:25][:25]
predicted_lab = y_test_pred[:25][:25]

fig, ax = plt.subplots(nrows=5, ncols=5, sharex=True, sharey=True,)
ax = ax.flatten()
for i in range(25):
    img = orign_img[i].reshape(28, 28)
    ax[i].imshow(img, cmap='Greys', interpolation='nearest')
    ax[i].set_title('%d) t: %d p: %d' % (i+1, true_lab[i], predicted_lab[i]))

ax[0].set_xticks([])
ax[0].set_yticks([])
plt.show()

## 逆正則化パラメータcを変化させたときのtraining, testデータに対するaccuracy，
## および非ゼロの重みの数を保存する．
weights, params = [], []
n_nonzero_weights, accuracy_train, accuracy_test = [], [], [] 
for c in np.arange(-11, 11, dtype=np.float):
    lr = LogisticRegression(penalty='l1', C=10**c, random_state=0)
    lr.fit(X_train, y_train)
    weights.append(lr.coef_[1])
    n_nonzero_weights.append(np.count_nonzero(lr.coef_[1]))
    params.append(10**c)
    y_train_pred = lr.predict(X_train)
    y_test_pred = lr.predict(X_test)
    acc_train_temp = np.sum(y_train == y_train_pred, axis=0) *100 / X_train.shape[0] 
    acc_test_temp = np.sum(y_test == y_test_pred, axis=0) *100 / X_test.shape[0] 
    accuracy_train.append(acc_train_temp) 
    accuracy_test.append(acc_test_temp) 

weights = np.array(weights)
n_nonzero_weights = np.array(n_nonzero_weights)

# 画像中心付近の2点に対する重みの変化をプロット
plt.figure(2)
# Feature from pixel row 15, col 10
plt.plot(params, weights[:, 402],
         label='Feature #402 (row 15, col 10)')
# Feature from pixel row 15, col 13
plt.plot(params, weights[:, 405], linestyle='--',
         label='Feature #405 (row 15, col 13)')
plt.ylabel('weight coefficient')
plt.xlabel('C')
plt.legend(loc='upper left')
plt.xscale('log')
plt.show()

# 保存した逆正則化パラメータcとAccuracy，および非ゼロの重みの数をグラフにプロットする．
plt.figure(3)
accuracy_train = np.array(accuracy_train)
accuracy_test = np.array(accuracy_test)
plt.plot(params, accuracy_train[:],label='Training')
plt.plot(params, accuracy_test[:],label='Testing')
print(accuracy_train[:])
plt.ylabel('Accuracy')
plt.xlabel('C')
plt.legend(loc='upper left')
plt.xscale('log')

plt.figure(4)
plt.plot(params, n_nonzero_weights)
plt.ylabel('# non-zero weights')
plt.xlabel('C')
plt.xscale('log')
plt.show()