from keras.utils import np_utils
from keras.models import Sequential, Model
from keras.layers import Activation, Dense, Dropout, Input
from keras.optimizers import Adam
import matplotlib.pyplot as plt
import os, struct
import numpy as np

# =====================================================================
# MNISTデータの読み込み関数
def load_mnist(path, kind='train'):

    labels_path = os.path.join(path,'%s-labels-idx1-ubyte'% kind)
    images_path = os.path.join(path,'%s-images-idx3-ubyte'% kind)
   
    with open(labels_path, 'rb') as lbpath:
        magic, n = struct.unpack('>II',lbpath.read(8))
        labels = np.fromfile(lbpath,dtype=np.uint8)
    
    with open(images_path, 'rb') as imgpath:
        magic, num, rows, cols = struct.unpack(">IIII",imgpath.read(16))
        images = np.fromfile(imgpath,dtype=np.uint8).reshape(len(labels), 784)
        
    return images, labels

# MNISTデータの読み込み
current_path = os.path.dirname(os.path.realpath(__file__))
X_train, y_train = load_mnist(current_path, kind='train')
X_test, y_test = load_mnist(current_path, kind='t10k')

# =====================================================================
# 学習データとテストデータは最初のn_train_data, n_test_data個用いる
n_training_data = 1000
n_test_data = 1000

X_trn = X_train[:n_training_data][:]
y_trn = y_train[:n_training_data]
X_tst = X_test[:n_test_data][:]
y_tst = y_test[:n_test_data]

# 値の範囲を[0,1]に変換
X_trn = X_trn.astype('float32')/255
X_tst = X_tst.astype('float32')/255

# One-hot encoderによりクラスラベルをバイナリに変換
# 例：1 -> [0,1,0,...,0], 2 -> [0,0,1,0,...]
y_trn = np_utils.to_categorical(y_trn)
y_tst = np_utils.to_categorical(y_tst)

# 入力データの次元数(=784画素)を取得
n_dim = X_trn.shape[1]

# 出力クラス数(=10クラス)
n_out = y_trn.shape[1]

# =====================================================================
# 学習履歴をプロットする関数

# Accuracyの履歴のプロット
def plot_history_acc(rec):
    plt.plot(rec.history['acc'],"o-",label="train")
    plt.plot(rec.history['val_acc'],"o-",label="test")
    plt.title('accuracy history')
    plt.xlabel('epochs')
    plt.ylabel('accuracy')
    plt.legend(loc="lower right")
    plt.show()
    
# 損失関数値の履歴のプロット
def plot_history_loss(rec):
    plt.plot(rec.history['loss'],"o-",label="train",)
    plt.plot(rec.history['val_loss'],"o-",label="test")
    plt.title('loss history')
    plt.xlabel('epochs')
    plt.ylabel('loss')
    plt.legend(loc='upper right')
    plt.show()

# =====================================================================
# AutoEncoderの構築

ae = Sequential()
ae.add(Dense(500, input_dim = n_dim, activation='relu'))
ae.add(Dropout(0.2))
ae.add(Dense(250, activation='relu'))
ae.add(Dropout(0.5))
ae.add(Dense(125, activation='relu', name = 'encoder'))
ae.add(Dropout(0.5))
ae.add(Dense(250, activation='relu'))
ae.add(Dropout(0.5))
ae.add(Dense(500, activation='relu'))
ae.add(Dropout(0.5))
ae.add(Dense(n_dim, activation='relu'))

ae.compile(loss = 'mse', optimizer ='adam')
records_ae = ae.fit(X_trn, X_trn,
                    epochs=250,
                    batch_size=200,
                    shuffle=True,
                    validation_data=(X_tst, X_tst))

# 学習済み重みの保存
ae.save_weights('autoencoder.h5')
# ネットワークの概要
ae.summary()
# 損失関数値の履歴のプロット
plot_history_loss(records_ae)

# =====================================================================
# AutoEncoderで再構成した画像の表示
def plot_reconstructed_images():
    # テスト画像をAutoEncoderで変換
    decoded_imgs = ae.predict(X_tst)

    n = 10 #表示枚数
    plt.figure(figsize=(20, 4))
    for i in range(n):
    	# 元画像の表示
        ax = plt.subplot(2, n, i+1)
        plt.imshow(X_tst[i].reshape(28, 28))
        plt.gray()
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        # 再構成画像の表示
        ax = plt.subplot(2, n, i+1+n)
        plt.imshow(decoded_imgs[i].reshape(28, 28))
        plt.gray()
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
    plt.show()

# 関数を呼び出して再構成画像との比較表示
plot_reconstructed_images()

# =====================================================================
# AutoEncoderの学習結果を用いて深層ニューラルネットを構成（DNN）

# AutoEncoderの学習結果（Encoderの学習済み重み）を取得
h = ae.get_layer('encoder').output
# 最終段にクラス数の分の出力を持つsoftmax関数を追加
y = Dense(n_out, activation='softmax', name='predictions')(h)

dnn = Model(inputs=ae.inputs, outputs=y)
dnn.compile(optimizer='adam', loss='categorical_crossentropy',metrics=['accuracy'])

records_dnn = dnn.fit(X_trn, y_trn,
		epochs=50,
                batch_size=200,
                shuffle=True,
                validation_data=(X_tst, y_tst))

# ネットワークの概要
dnn.summary()
# 学習履歴のプロット
plot_history_acc(records_dnn)
plot_history_loss(records_dnn)

# =====================================================================
# AutoEncoderを使わない中間層1層の多層パーセプトロン（MLP）

mlp = Sequential()
mlp.add(Dense(500, input_dim = n_dim, activation='sigmoid'))
mlp.add(Dense(n_out, activation='softmax'))
mlp.compile(loss = 'categorical_crossentropy', optimizer ='adam', metrics = ['accuracy'])

records_mlp = mlp.fit(X_trn, y_trn,
                epochs=100,
                batch_size=200,
                validation_data=(X_tst, y_tst))
                
# ネットワークの概要
mlp.summary()

# =====================================================================
# Accuracyの比較
loss_dnn, acc_dnn = dnn.evaluate(X_tst, y_tst, verbose=0)
loss_mlp, acc_mlp = mlp.evaluate(X_tst, y_tst, verbose=0)
print('===========')
print('Test accuracy (DNN):', acc_dnn)
print('Test accuracy (MLP):', acc_mlp)
print('===========')
